"""Private optional import utils.

Not for use outside trulens namespace.
"""

from trulens.core.utils import imports as import_utils

# Optional app types:
REQUIREMENT_APPS_LLAMA = import_utils.format_import_errors(
    "trulens-apps-llamaindex", purpose="instrumenting LlamaIndex apps"
)
REQUIREMENT_APPS_LANGCHAIN = import_utils.format_import_errors(
    "trulens-apps-langchain", purpose="instrumenting LangChain apps"
)
REQUIREMENT_APPS_NEMO = import_utils.format_import_errors(
    "trulens-apps-nemo", purpose="instrumenting NeMo Guardrails apps"
)

REQUIREMENT_TQDM = import_utils.format_import_errors(
    "tqdm", purpose="displaying progress bars"
)



import asyncio
from collections import deque
from datetime import datetime
from datetime import timedelta
import logging
from threading import Lock
import time
from typing import ClassVar, Deque, Optional

from pydantic import BaseModel
from pydantic import Field
from pydantic import PrivateAttr

logger = logging.getLogger(__name__)


class Pace(BaseModel):
    """Keep a given pace.

    Calls to `Pace.mark` may block until the pace of its returns is kept to a
    constraint: the number of returns in the given period of time cannot exceed
    `marks_per_second * seconds_per_period`. This means the average number of
    returns in that period is bounded above exactly by `marks_per_second`.

    !!! Warning:
        The asynchronous and synchronous methods `amark` and `mark` should not be
        used at the same time. That is, use either the synchronous interface or the
        asynchronous one, but not both.
    """

    marks_per_second: float = 1.0
    """The pace in number of mark returns per second."""

    seconds_per_period: float = 60.0
    """Evaluate pace as the average over this period.

    Assumes that prior to construction of this Pace instance, the period did not
    have any marks called. The longer this period is, the bigger burst of marks
    will be allowed initially and after long periods of no marks.
    """

    seconds_per_period_timedelta: timedelta = Field(
        default_factory=lambda: timedelta(seconds=60.0)
    )
    """The above period as a timedelta."""

    mark_expirations: Deque[datetime] = Field(default_factory=deque)
    """Keep track of returns that happened in the last `period` seconds.

    Store the datetime at which they expire (they become older than `period`
    seconds old).
    """

    max_marks: int
    """The maximum number of marks to keep track in the above deque.

    It is set to (seconds_per_period * returns_per_second) so that the
    average returns per second over period is no more than exactly
    returns_per_second.
    """

    last_mark: datetime = Field(default_factory=datetime.now)
    """Time of the last mark return."""

    _lock: Lock = PrivateAttr(default_factory=Lock)
    """Thread Lock to ensure mark method details run only one at a time."""

    _alock: asyncio.Lock = PrivateAttr(default_factory=asyncio.Lock)
    """Asyncio Lock to ensure amark method details run only one at a time."""

    model_config: ClassVar[dict] = dict(arbitrary_types_allowed=True)

    _warned: bool = False
    """Whether the long delay warning has already been issued.

    This is to not repeatedly give it.
    """

    def __init__(
        self,
        seconds_per_period: float,
        *args,
        marks_per_second: Optional[float] = None,
        rpm: Optional[float] = None,
        **kwargs,
    ):
        if marks_per_second is None:
            assert (
                rpm is not None
            ), "Either `marks_per_second` or `rpm` must be given."
            marks_per_second = rpm / 60.0
        else:
            assert (
                rpm is None
            ), "Only one of `marks_per_second` or `rpm` can be given."

        max_marks = int(seconds_per_period * marks_per_second)
        if max_marks == 0:
            raise ValueError(
                "Period is too short for the give rate. "
                "Increase `seconds_per_period` or `marks_per_second` (or both)."
            )

        self._ensure_event_loop_exists()

        super().__init__(
            *args,
            seconds_per_period=seconds_per_period,
            seconds_per_period_timedelta=timedelta(seconds=seconds_per_period),
            marks_per_second=marks_per_second,
            max_marks=max_marks,
            **kwargs,
        )

    @staticmethod
    def _ensure_event_loop_exists():
        try:
            asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

    async def amark(self) -> float:
        """Return in appropriate pace.

        Blocks until return can happen in the appropriate pace. Returns time in
        seconds since last mark returned.
        """

        async with self._alock:
            while len(self.mark_expirations) >= self.max_marks:
                delay = (
                    self.mark_expirations[0] - datetime.now()
                ).total_seconds()

                if delay >= self.seconds_per_period * 0.5:
                    if not self._warned:
                        self._warned = True
                        logger.warning(
                            """
Pace has a long delay of %s seconds. There might have been a burst of
requests which may become a problem for the receiver of whatever is being paced.
Consider reducing the `seconds_per_period` (currently %s [seconds]) over which to
maintain pace to reduce burstiness. " Alternatively reduce `marks_per_second`
(currently %s [1/second]) to reduce the number of marks
per second in that period.
    """,
                            delay,
                            self.seconds_per_period,
                            self.marks_per_second,
                        )

                if delay > 0.0:
                    await asyncio.sleep(delay)

                self.mark_expirations.popleft()

            prior_last_mark = self.last_mark
            now = datetime.now()
            self.last_mark = now

            # Add to marks the point at which the mark can be removed (after
            # `period` seconds).
            self.mark_expirations.append(
                now + self.seconds_per_period_timedelta
            )

            return (now - prior_last_mark).total_seconds()

    def mark(self) -> float:
        """Return in appropriate pace.

        Blocks until return can happen in the appropriate pace. Returns time in
        seconds since last mark returned.
        """

        with self._lock:
            while len(self.mark_expirations) >= self.max_marks:
                delay = (
                    self.mark_expirations[0] - datetime.now()
                ).total_seconds()

                if delay >= self.seconds_per_period * 0.5:
                    if not self._warned:
                        self._warned = True
                        logger.warning(
                            """
Pace has a long delay of %s seconds. There might have been a burst of
requests which may become a problem for the receiver of whatever is being paced.
Consider reducing the `seconds_per_period` (currently %s [seconds]) over which to
maintain pace to reduce burstiness. " Alternatively reduce `marks_per_second`
(currently %s [1/second]) to reduce the number of marks
per second in that period.
""",
                            delay,
                            self.seconds_per_period,
                            self.marks_per_second,
                        )

                if delay > 0.0:
                    time.sleep(delay)

                self.mark_expirations.popleft()

            prior_last_mark = self.last_mark
            now = datetime.now()
            self.last_mark = now

            # Add to marks the point at which the mark can be removed (after
            # `period` seconds).
            self.mark_expirations.append(
                now + self.seconds_per_period_timedelta
            )

            return (now - prior_last_mark).total_seconds()



# WARNING: This file does not follow the no-init aliases import standard.

from trulens.core.feedback.endpoint import Endpoint
from trulens.core.feedback.endpoint import EndpointCallback
from trulens.core.feedback.feedback import Feedback
from trulens.core.feedback.feedback import SkipEval
from trulens.core.feedback.feedback import SnowflakeFeedback
from trulens.core.feedback.provider import Provider

__all__ = [
    "Feedback",
    "SnowflakeFeedback",
    "Provider",
    "Endpoint",
    "EndpointCallback",
    "SkipEval",
]



"""Add run_location column to feedback_defs table.

Revision ID: 2
Revises: 1
Create Date: 2024-08-15 15:43:22.560492
"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "2"
down_revision = "1"
branch_labels = None
depends_on = "1"


def upgrade(config) -> None:
    prefix = config.get_main_option("trulens.table_prefix")

    if prefix is None:
        raise RuntimeError("trulens.table_prefix is not set")

    # ### begin Alembic commands ###
    op.add_column(
        prefix + "feedback_defs",
        sa.Column("run_location", sa.Text(), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade(config) -> None:
    prefix = config.get_main_option("trulens.table_prefix")

    if prefix is None:
        raise RuntimeError("trulens.table_prefix is not set")

    # ### begin Alembic commands ###
    op.drop_column(prefix + "feedback_defs", "run_location")
    # ### end Alembic commands ###



"""
Container class utilities.
"""

from __future__ import annotations

import datetime
import itertools
import logging
from pprint import PrettyPrinter
from threading import Event
from threading import RLock
from typing import (
    Callable,
    Dict,
    Generic,
    Iterable,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
)

import pandas as pd

logger = logging.getLogger(__name__)
pp = PrettyPrinter()

T = TypeVar("T")
A = TypeVar("A")
B = TypeVar("B")


def datetime_of_ns_timestamp(timestamp: int) -> datetime.datetime:
    """Convert a nanosecond timestamp to a datetime."""
    return pd.Timestamp(timestamp, unit="ns").to_pydatetime()


def ns_timestamp_of_datetime(dt: datetime.datetime) -> int:
    """Convert a datetime to a nanosecond timestamp."""
    return pd.Timestamp(dt).as_unit("ns").value


class BlockingSet(set, Generic[T]):
    """A set with max size that has blocking peek/get/add ."""

    def __init__(
        self, items: Optional[Iterable[T]] = None, max_size: int = 1024
    ):
        if items is not None:
            items = list(items)
        else:
            items = []

        if len(items) > max_size:
            raise ValueError("Initial items exceed max size.")

        self.content: Set[T] = set(items)
        self.max_size = max_size

        # TODO: unsure if these 2 locks are sufficient to prevent all deadlocks
        self.read_lock = RLock()
        self.write_lock = RLock()
        self.event_nonempty = Event()
        self.event_nonfull = Event()
        self.event_shutdown = Event()

        if len(self.content) > 0:
            self.event_nonempty.set()

        if len(self.content) < self.max_size:
            self.event_nonfull.set()

    def empty(self) -> bool:
        """Check if the set is empty."""
        return len(self.content) == 0

    def __del__(self):
        self.shutdown()

    def shutdown(self):
        """Shutdown the set."""

        self.event_shutdown.set()
        self.event_nonempty.set()  # Unblock any waiting threads

    def peek(self) -> T:
        """Get an item from the set.

        Blocks until an item is available.
        """

        with self.read_lock:
            self.event_nonempty.wait()
            if self.event_shutdown.is_set():
                raise StopIteration("Set is shutdown.")

            return next(iter(self.content))

    def remove(self, item: T):
        """Remove an item from the set."""
        with self.write_lock:
            self.content.remove(item)

            self.event_nonfull.set()

            if len(self.content) == 0:
                self.event_nonempty.clear()

    def pop(self, blocking: bool = True) -> Optional[T]:
        """Get and remove an item from the set.

        Blocks until an item is available, unless blocking is set to False.

        Args:
            blocking: Whether to block until an item is ready. If not blocking
                and empty, will return None.
        """

        if self.read_lock.acquire(blocking=blocking):
            if not blocking and not self.event_nonempty.is_set():
                self.read_lock.release()
                return None

            self.event_nonempty.wait()

            if self.event_shutdown.is_set():
                self.read_lock.release()
                raise StopIteration("Set is shutdown.")

            item = next(iter(self.content))
            self.content.remove(item)

            if len(self.content) == 0:
                self.event_nonempty.clear()

            self.read_lock.release()

        else:
            # Acquire was used with blocking=False and it would have
            # blocked. Want the non-blocking behaviour to return None in such cases.
            return None

        return item

    def add(self, item: T):
        """Add an item to the set.

        Blocks if set is full.
        """

        with self.write_lock:
            self.event_nonfull.wait()
            self.content.add(item)

            self.event_nonempty.set()
            if len(self.content) >= self.max_size:
                self.event_nonfull.clear()


# Collection utilities


def first(seq: Sequence[T]) -> T:
    """Get the first item in a sequence."""
    return seq[0]


def second(seq: Sequence[T]) -> T:
    """Get the second item in a sequence."""
    return seq[1]


def third(seq: Sequence[T]) -> T:
    """Get the third item in a sequence."""
    return seq[2]


def is_empty(obj):
    """Check if an object is empty.

    If object is not a sequence, returns False.
    """

    try:
        return len(obj) == 0
    except Exception:
        return False


def dict_set_with(dict1: Dict[A, B], dict2: Dict[A, B]) -> Dict[A, B]:
    """Add the key/values from `dict2` to `dict1`.

    Mutates and returns `dict1`.
    """

    dict1.update(dict2)
    return dict1


def dict_set_with_multikey(
    dict1: Dict[A, B], dict2: Dict[Union[A, Tuple[A, ...]], B]
) -> Dict[A, B]:
    """Like `dict_set_with` except the second dict can have tuples as keys in which
    case all of the listed keys are set to the given value."""
    for ks, v in dict2.items():
        if isinstance(ks, tuple):
            for k in ks:
                dict1[k] = v
        else:
            dict1[ks] = v
    return dict1


def dict_merge_with(dict1: Dict, dict2: Dict, merge: Callable) -> Dict:
    """Merge values from the second dictionary into the first.

    If both dicts contain the same key, the given `merge` function is used to
    merge the values.
    """
    for k, v in dict2.items():
        if k in dict1:
            dict1[k] = merge(dict1[k], v)
        else:
            dict1[k] = v

    return dict1


# Generator utils


def iterable_peek(it: Iterable[T]) -> Tuple[T, Iterable[T]]:
    iterator = iter(it)
    item = next(iterator)
    return item, itertools.chain([item], iterator)



"""Common/shared serializable classes."""

from __future__ import annotations

import datetime
from typing import Optional

import pydantic
from trulens.core.utils import serial as serial_utils

MAX_DILL_SIZE: int = 1024 * 1024  # 1MB
"""Max size in bytes of pickled objects."""


class Cost(serial_utils.SerialModel, pydantic.BaseModel):
    """Costs associated with some call or set of calls."""

    n_requests: int = 0
    """Number of requests.

    To increment immediately when a request is made.
    """

    n_successful_requests: int = 0
    """Number of successful requests.

    To increment only after a response is processed and it indicates success.
    """

    n_completion_requests: int = 0
    """Number of completion requests.

    To increment immediately when a completion request is made.
    """

    n_classification_requests: int = 0
    """Number of classification requests.

    To increment immediately when a classification request is made.
    """

    n_classes: int = 0
    """Number of class scores retrieved.

    To increment for each class in a successful classification response.
    """

    n_embedding_requests: int = 0
    """Number of embedding requests.

    To increment immediately when an embedding request is made.
    """

    n_embeddings: int = 0
    """Number of embeddings retrieved.

    To increment for each embedding vector returned by an embedding request.
    """

    n_tokens: int = 0
    """Total tokens processed.

    To increment by the number of input(prompt) and output tokens in completion
    requests. While the input part of this could be incremented upon a request,
    the actual count is not easy to determine due to tokenizer variations and
    instead is usually seen in the response. Also, we want to count only tokens
    for successful requests that incur a cost to the user.
    """

    n_stream_chunks: int = 0
    """In streaming mode, number of chunks produced.

    To increment for each chunk in a streaming response. This does not need to
    wait for completion of the responses.
    """

    n_prompt_tokens: int = 0
    """Number of prompt tokens supplied.

    To increment by the number of tokens in the prompt of a completion request.
    This is visible in the response though and should only count successful
    requests.
    """

    n_completion_tokens: int = 0
    """Number of completion tokens generated.

    To increment by the number of tokens in the completion of a completion request.
    """

    n_cortex_guardrails_tokens: int = 0
    """Number of guardrails tokens generated. This is only available for
    requests instrumented by the Cortex endpoint."""

    cost: float = 0.0
    """Cost in [cost_currency].

    This may not always be available or accurate.
    """

    cost_currency: str = "USD"

    def __add__(self, other: "Cost") -> "Cost":
        kwargs = {
            k: getattr(self, k) + getattr(other, k)
            if k != "cost_currency"
            and isinstance(getattr(self, k), (int, float))
            else getattr(other, k)
            for k in self.model_fields.keys()
        }
        if other.cost_currency != self.cost_currency:
            if self.cost == 0:
                kwargs["cost_currency"] = other.cost_currency
            elif other.cost == 0:
                kwargs["cost_currency"] = self.cost_currency
            else:
                raise ValueError(
                    f"Cannot add costs with different currencies: {self.cost_currency} and {other.cost_currency}!"
                )
        return Cost(**kwargs)

    def __radd__(self, other: "Cost") -> "Cost":
        # Makes sum work on lists of Cost.

        if other == 0:
            return self

        return self.__add__(other)


class Perf(serial_utils.SerialModel, pydantic.BaseModel):
    """Performance information.

    Presently only the start and end times, and thus latency.
    """

    start_time: datetime.datetime
    """Datetime before the recorded call."""

    end_time: datetime.datetime
    """Datetime after the recorded call."""

    @staticmethod
    def min():
        """Zero-length span with start and end times at the minimum datetime."""

        return Perf(
            start_time=datetime.datetime.min, end_time=datetime.datetime.min
        )

    @staticmethod
    def now(latency: Optional[datetime.timedelta] = None) -> Perf:
        """Create a `Perf` instance starting now and ending now plus latency.

        Args:
            latency: Latency in seconds. If given, end time will be now plus
                latency. Otherwise end time will be a minimal interval plus start_time.
        """

        start_time = datetime.datetime.now()
        if latency is not None:
            end_time = start_time + latency
        else:
            end_time = start_time + datetime.timedelta(microseconds=1)

        return Perf(start_time=start_time, end_time=end_time)

    @property
    def latency(self):
        """Latency in seconds."""
        return self.end_time - self.start_time


# HACK013: Need these if using __future__.annotations .
Cost.model_rebuild()
Perf.model_rebuild()



"""Utilities related to wrapping live python objects."""

from __future__ import annotations

import functools
import inspect
import logging
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    Optional,
    Tuple,
    Type,
    TypeVar,
)
import uuid

from trulens.core.utils import python as python_utils

logger = logging.getLogger(__name__)

R = TypeVar("R")  # callable's return type
A = TypeVar("A")  # awaitable's result type
E = TypeVar("E")  # iterator/generator element type


class AwaitableCallbacks(Generic[A]):
    """Callbacks for wrapped awaitables.

    This class is intentionally not an ABC so as to make default callbacks
    no-ops.
    """

    @staticmethod
    def on_awaitable_wrapped(
        awaitable: Awaitable[A], wrapper: Awaitable[A], **kwargs
    ) -> None:
        """Called immediately after the wrapper is produced."""

    def __init__(
        self,
        awaitable: Awaitable[A],
        wrapper: Awaitable[A],
        **kwargs: Dict[str, Any],
    ):
        """Called/constructed when wrapper awaitable is awaited but before the
        wrapped awaitable is awaited."""

        # Subclasses can use kwargs.

        self.awaitable: Awaitable[A] = awaitable
        self.wrapper: Awaitable[A] = wrapper

        self.result: Optional[A] = None
        self.error: Optional[Exception] = None

    def on_awaitable_end(self):
        """Called after the last callback is called."""

    def on_awaitable_result(self, result: A) -> A:
        """Called with the result of the wrapped awaitable once it is ready.

        !!! Important
            This should return the result or some wrapper of the result.
        """
        self.result = result
        return result

    def on_awaitable_exception(self, error: Exception) -> Exception:
        """Called if awaiting for the wrapped awaitable raised an exception."""
        self.error = error
        return error


class AwaitableCallbacksFromCallableCallbacks(AwaitableCallbacks[A]):
    """Wraps a CallableCallbacks instance to provide AwaitableCallbacks.

    This assumes that the input instances was the one that handled the wrapping
    of the function that produced the awaitable result.

    Callbacks are invoked like this:

    - on_callable_return with on_awaitable_result with ret as result,

    - on_callable_end with on_awaitable_end,

    - on_callable_exception with on_awaitable_exception.

    - on_callable_call is not used as it was assumed to have already been
      called.
    """

    def __init__(
        self,
        awaitable: Awaitable[A],
        wrapper: Awaitable[A],
        callbacks: CallableCallbacks[A],
        **kwargs: Dict[str, Any],
    ):
        super().__init__(awaitable, wrapper, **kwargs)
        self._callable_callbacks = callbacks

    def on_awaitable_end(self):
        return self._callable_callbacks.on_callable_end()

    def on_awaitable_result(self, result: A) -> A:
        return self._callable_callbacks.on_callable_return(ret=result)

    def on_awaitable_exception(self, error: Exception) -> Exception:
        return self._callable_callbacks.on_callable_exception(error=error)


def wrap_awaitable(
    awaitable: Awaitable[A],
    callback_class: Type[AwaitableCallbacks[A]] = AwaitableCallbacks,
    **kwargs: Dict[str, Any],
) -> Awaitable[A]:
    """Wrap an awaitable in another awaitable that will call callbacks before
    and after the given awaitable finishes.

    Note that the resulting awaitable needs to be awaited for all but the
    `on_now` callback to trigger.

    Args:
        awaitable: The awaitable to wrap.

        callback_class: The class that provides callbacks.

        **kwargs: All other arguments are passed to the callback class
            constructor and to `on_awaitable_wrapped`.
    """

    init_args: Dict[str, Any] = {"awaitable": awaitable, **kwargs}

    async def wrapper(awaitable: Awaitable[A]) -> A:
        cb: AwaitableCallbacks[A] = callback_class(**init_args)

        try:
            result = await awaitable
            result = cb.on_awaitable_result(result=result)
            cb.on_awaitable_end()
            return result

        except Exception as e:
            e_wrapped = cb.on_awaitable_exception(error=e)
            cb.on_awaitable_end()

            if e == e_wrapped:
                raise e

            raise e_wrapped from e

    w: Awaitable[A] = wrapper(awaitable)

    init_args["wrapper"] = w
    callback_class.on_awaitable_wrapped(**init_args)

    return w


class IterableCallbacks(Generic[E]):
    """Callbacks for wrapped iterables.

    This class is intentionally not an ABC so as to make default callbacks
    no-ops.
    """

    @staticmethod
    def on_iterable_wrapped(
        itb: Iterable[E], wrapper: Iterable[E], **kwargs: Dict[str, Any]
    ):
        """Called immediately after wrapper is made."""

    def __init__(
        self, itb: Iterable[E], wrapper: Iterable[E], **kwargs: Dict[str, Any]
    ):
        """Called/constructed when the wrapped iterable is iterated (__iter__ is
        called) but before the wrapped iterable is iterated (__iter__ is
        called)."""

        # Subclasses can use kwargs.

        self.itb: Iterable[E] = itb
        self.wrapper: Iterable[E] = wrapper

        self.it: Optional[Iterator[E]] = None
        self.error: Optional[Exception] = None

    def on_iterable_end(self):
        """Called after the last callback is called."""

    def on_iterable_iter(
        self,
        it: Iterator[E],
    ) -> Iterator[E]:
        """Called after the wrapped iterator is produced (but before any items
        are produced).

        !!! Important
            Must produce the given iterator it or some wrapper.
        """

        self.it = it
        return it

    def on_iterable_iter_exception(self, error: Exception) -> Exception:
        """Called if an error is raised during __iter__."""

        self.error = error
        return error

    def on_iterable_next(self):
        """Called when the next item from the iterator is requested (but before
        it is produced)."""

    def on_iterable_next_result(self, val: E) -> E:
        """Called after the next item from the iterator is produced.

        !!! Important
            Must produce the given value val or some wrapper.
        """
        return val

    def on_iterable_next_exception(self, error: Exception):
        """Called if an error is raised during __next__."""
        self.error = error
        return error

    def on_iterable_stop(self):
        """Called after the iterator stops iterating (raises StopIteration)."""


def wrap_iterable(
    itb: Iterable[E],
    callback_class: Type[IterableCallbacks] = IterableCallbacks,
    **kwargs: Dict[str, Any],
) -> Iterable[E]:
    """Wrap an iterable to invoke various callbacks.

    See
    [IterableCallbacks][trulens.experimental.otel_tracing.core._utils.wrap.IterableCallbacks]
    for callbacks invoked.

    Args:
        itb: The iterator to wrap.

        **kwargs: All other arguments are passed to the callback class
            constructor and to `on_iterable_wrapped`,
            `on_iterable_iter_exception`.
    """

    init_args = {"itb": itb, **kwargs}

    def wrapper(itb: Iterable[E]) -> Iterable[E]:
        cb = callback_class(**init_args)

        try:
            it: Iterator[E] = iter(itb)
            it = cb.on_iterable_iter(it=it)

        except Exception as e:
            init_args["error"] = e
            e_wrapped = cb.on_iterable_iter_exception(**init_args)
            cb.on_iterable_end()

            if e == e_wrapped:
                raise e

            raise e_wrapped from e

        while True:
            cb.on_iterable_next()

            try:
                val = next(it)
                val = cb.on_iterable_next_result(val=val)
                yield val

            except StopIteration:
                cb.on_iterable_stop()
                cb.on_iterable_end()
                return

            except Exception as e:
                e_wrapped = cb.on_iterable_next_exception(error=e)
                cb.on_iterable_end()

                if e == e_wrapped:
                    raise e

                raise e_wrapped from e

    w: Iterable[E] = wrapper(itb)

    init_args["wrapper"] = w
    callback_class.on_iterable_wrapped(**init_args)

    return w


class CallableCallbacks(Generic[R]):
    """Callbacks for wrapped callables.

    This class is intentionally not an ABC so as to make default callbacks
    no-ops.
    """

    @classmethod
    def on_callable_wrapped(
        cls,
        *,
        func: Callable[..., R],
        wrapper: Callable[..., R],
        **kwargs: Dict[str, Any],
    ):
        """Called immediately after the wrapper is produced.

        !!! NOTE
            There is only one wrapper produced for each wrapped function. If
            wrapping is done again for an already wrapped function, this
            callback will get called with the existing wrapper.
        """

    def __init__(
        self,
        *,
        call_id: uuid.UUID,
        func: Callable[..., R],
        sig: inspect.Signature,
        wrapper: Callable[..., R],
        call_args: Tuple[Any, ...],
        call_kwargs: Dict[str, Any],
    ):
        """Called/constructed when the wrapper function is called but before
        arguments are bound to the wrapped function's signature."""

        # Subclasses can use kwargs.

        self.call_id: uuid.UUID = call_id
        self.func: Callable[..., R] = func
        self.sig: inspect.Signature = sig
        self.wrapper: Callable[..., R] = wrapper
        self.call_args: Optional[Tuple[Any, ...]] = call_args
        self.call_kwargs: Optional[Dict[str, Any]] = call_kwargs

        self.bindings: Optional[inspect.BoundArguments] = None
        self.bind_error: Optional[Exception] = None

        self.error: Optional[Exception] = None

        self.ret: Optional[R] = None

    def on_callable_end(self):
        """Called after the last callback is called."""

    def on_callable_call(
        self,
        *,
        bindings: inspect.BoundArguments,
    ) -> inspect.BoundArguments:
        """Called before the execution of the wrapped method assuming its
        arguments can be bound.

        !!! Important
            This callback must return the bound arguments or wrappers of bound
            arguments.
        """
        self.bindings = bindings

        return bindings

    def on_callable_bind_error(
        self,
        *,
        error: TypeError,
    ):
        """Called if the wrapped method's arguments cannot be bound.

        Note that if this type of error occurs, none of the wrapped code
        actually executes. Also if this happens, the `on_callable_exception`
        handler is expected to be called as well after the wrapped func is
        called with the unbindable arguments. This is done to replicate the
        error behavior of an unwrapped invocation.
        """

        self.bind_error = error

    def on_callable_return(self, *, ret: R) -> R:
        """Called after wrapped method returns without error.

        !!! Important
            This callback must return the return value or some wrapper of that
            value.
        """
        self.ret = ret

        return ret

    def on_callable_exception(self, *, error: Exception) -> Exception:
        """Called after wrapped method raises exception."""

        self.error = error
        return error


CALLBACKS = "__tru_callbacks"


def wrap_callable(
    func: Callable[..., R],
    callback_class: Type[CallableCallbacks[R]],
    **kwargs: Dict[str, Any],
) -> Callable[..., R]:
    """Create a wrapper of the given function to emit various callbacks at
    different stages of its evaluation.

    !!! Warning
        Note that we only allow a single wrapper and a callback class to be used
        at once. Managing multiple receivers of the data collected must be
        managed outside of this wrapping utility.

    Args:
        func: The function to wrap.

        callback_class: The class that provides callbacks.

        **kwargs: All other arguments are passed to the
            callback class constructor and to `on_callable_wrapped`.
    """

    if hasattr(func, "__wrapper__"):
        # For cases where multiple refs to func are captured before we we begin
        # wrapping it and each of the refs is wrapped. We don't want to re-wrap
        # as we would miss some callbacks. We thus store the wrapper in
        # func.__wrapper__ and check for it here. The rest of the call will
        # assume func is the wrapper and will add more callbacks instead of
        # creating a new wrapper.
        func = func.__wrapper__

    assert isinstance(func, Callable), f"Callable expected but got {func}."

    if python_utils.safe_hasattr(func, CALLBACKS):
        # If CALLBACKS is set, it was already a wrapper.

        # logger.warning("Function %s is already wrapped.", func)

        return func

    sig = inspect.signature(func)  # safe sig?
    our_callback_init_kwargs: Dict[str, Any] = {"func": func, "sig": sig}

    # If CALLBACKS is not set, create a wrapper and return it.
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> R:
        logger.debug(
            "Calling instrumented method %s of type %s.", func, type(func)
        )

        call_id: uuid.UUID = uuid.uuid4()

        callback_class, other_callback_init_kwargs = getattr(wrapper, CALLBACKS)

        # callback step 0: create the callback instance
        callback = callback_class(
            call_id=call_id,
            call_args=args,
            call_kwargs=kwargs,
            **our_callback_init_kwargs,
            **other_callback_init_kwargs,
        )

        try:
            bindings = sig.bind(*args, **kwargs)  # save and reuse sig

            # callback step 1: call on_callable_call
            callback.on_callable_call(bindings=bindings)

        except TypeError as e:
            # callback step 2a: call on_callable_bind_error
            callback.on_callable_bind_error(error=e)

            # NOTE: Not raising e here to make sure the exception raised by the
            # wrapper is identical to the one produced by calling the wrapped
            # function below:

        try:
            # Get the result of the wrapped function.
            ret = func(*args, **kwargs)

            if isinstance(ret, Awaitable):
                # Arranges to call on_callable_return and on_callable_end after
                # the awaitable is done.

                # type R is same as Awaitable[A] for some type A

                return wrap_awaitable(
                    ret,
                    callback_class=AwaitableCallbacksFromCallableCallbacks,
                    callbacks=callback,
                )  # type: ignore[return-value]

            ret: R = callback.on_callable_return(ret=ret)  # type: ignore[no-redef]
            # Can override ret.

            callback.on_callable_end()
            return ret

        except Exception as e:
            # Exception on wrapped func call: ret = func(...)

            wrapped_e = callback.on_callable_exception(error=e)
            # Can override exception.

            callback.on_callable_end()

            raise wrapped_e  # intentionally not "from e" # pylint: disable=W0707

    # Set our tracking attribute to tell whether something is already
    # instrumented onto both the sync and async version since either one
    # could be returned from this method.

    our_callback_init_kwargs["wrapper"] = wrapper

    setattr(wrapper, CALLBACKS, (callback_class, kwargs))
    # The second item in the tuple are the other kwargs that were passed into
    # wrap_callable which are provided to the callback_class constructor.

    wrapper.__func__ = func
    func.__wrapper__ = wrapper

    callback_class.on_callable_wrapped(**our_callback_init_kwargs, **kwargs)

    return wrapper



from __future__ import annotations  # defers evaluation of annotations

from collections import defaultdict
from enum import Enum
import inspect
import json
import logging
import time
from typing import Any, ClassVar, Dict, List, Optional, Set, Type
import uuid

import pandas as pd
import pydantic
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field
from trulens.core.utils.json import obj_id_of_obj
from trulens.otel.semconv.trace import SpanAttributes

logger = logging.getLogger(__name__)


def _get_all_span_attribute_key_constants(cls: Type, prefix: str) -> List[str]:
    ret = []
    for curr_name in dir(cls):
        if (not curr_name.startswith("__")) and (not curr_name.endswith("__")):
            curr = getattr(cls, curr_name)
            if inspect.isclass(curr):
                ret += _get_all_span_attribute_key_constants(
                    curr, f"{curr_name}"
                )
            elif curr_name == curr_name.upper():
                ret += [f"{prefix + '.' if prefix else ''}{curr_name}"]
    ret = list(set(ret))
    ret = [field for field in ret if not (field.startswith("SpanType"))]

    ret += [
        field.replace(
            "RECORD_ROOT.", ""
        )  # record_root can be optionally specified
        for field in ret
        if field.startswith("RECORD_ROOT.")
    ]
    return ret


def get_all_span_attribute_key_constants() -> set[str]:
    return set(_get_all_span_attribute_key_constants(SpanAttributes, ""))


# Reserved fields (case-insensitive) for dataset specification directly maps to OTEL Span attributes
DATASET_RESERVED_FIELDS: Set[str] = {
    field.lower() for field in get_all_span_attribute_key_constants()
}


EXPECTED_TELEMETRY_LATENCY_IN_MS = (
    2 * 60 * 1000
)  # expected latency from the telemetry pipeline before ingested rows show up in event table


class RunStatus(str, Enum):
    # note this the inferred / computeted status determined by SDK, and is different from the DPO entity level run_status

    # invocation statuses
    INVOCATION_IN_PROGRESS = "INVOCATION_IN_PROGRESS"
    INVOCATION_COMPLETED = "INVOCATION_COMPLETED"
    INVOCATION_PARTIALLY_COMPLETED = "INVOCATION_PARTIALLY_COMPLETED"

    # computation statuses
    COMPUTATION_IN_PROGRESS = "COMPUTATION_IN_PROGRESS"
    COMPLETED = "COMPLETED"
    PARTIALLY_COMPLETED = "PARTIALLY_COMPLETED"

    CREATED = "CREATED"
    FAILED = "FAILED"
    CANCELLED = "CANCELLED"
    UNKNOWN = "UNKNOWN"


class SupportedEntryType(str, Enum):
    INVOCATIONS = "invocations"
    COMPUTATIONS = "computations"
    METRICS = "metrics"


SUPPORTED_ENTRY_TYPES = [e.value for e in SupportedEntryType]


def validate_dataset_spec(
    dataset_spec: Dict[str, str],
) -> Dict[str, str]:
    """
    Validates and normalizes the dataset column specification to ensure it contains only
    currently supported span attributes and that the keys are in the correct format.

    Args:
        dataset_spec: The user-provided dictionary with column names.

    Returns:
        A validated and normalized dictionary.

    Raises:
        ValueError: If any invalid field is present.
    """

    normalized_spec = {}

    for key, value in dataset_spec.items():
        normalized_key = key.lower()

        # Ensure that the key is one of the valid reserved fields or its subscripted form
        if not any(
            normalized_key.startswith(reserved_field)
            for reserved_field in DATASET_RESERVED_FIELDS
        ):
            raise ValueError(f"Invalid field '{key}' found in dataset_spec.")

        # Add the normalized field to the dictionary
        normalized_spec[normalized_key] = value

    return normalized_spec


class RunConfig(BaseModel):
    run_name: str = Field(
        ...,
        description="Unique name of the run. This name should be unique within the object.",
    )
    dataset_name: str = Field(
        default=...,
        description="Mandatory field. The name of a user's Snowflake Table / View  (e.g. 'user_table_name_1'), or any user specified name of input dataframe.",
    )

    source_type: str = Field(
        default="DATAFRAME",
        description="Type of the source (e.g. 'DATAFRAME' for user provided dataframe or 'TABLE' for user table in Snowflake).",
    )

    dataset_spec: Dict[str, str] = Field(
        default=...,
        description="Mandatory column name mapping from reserved dataset fields to column names in user's table.",
    )

    description: Optional[str] = Field(
        default=None, description="A description for the run."
    )
    label: Optional[str] = Field(
        default=None,
        description="Text label to group the runs. Take a single label for now",
    )
    llm_judge_name: Optional[str] = Field(
        default=None,
        description="Name of the LLM judge to be used for the run.",
    )


class Run(BaseModel):
    model_config: ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
        arbitrary_types_allowed=True,
        extra="ignore",  # allow custom obj like RunDao to be passed as a parameter and more importantly, account for
        # additional fields in Run metadata JSON response.
    )

    """
    Run class for managing run state / attributes in the SDK client.

    This model is meant to be used and accessed through
    methods like describe() (which uses the underlying RunDao) to obtain the run metadata.
    """

    run_dao: Any = Field(
        ..., description="DAO instance for run operations.", exclude=True
    )

    app: Any = Field(
        ...,
        description="TruLens app/recorder instance to be invoked during run.",
        exclude=True,
    )

    main_method_name: str = Field(
        ..., description="Main method of the app.", exclude=True
    )

    tru_session: Any = Field(
        ..., description="TruSession instance.", exclude=True
    )
    object_name: str = Field(
        ...,
        description="Name of the managing object (e.g. name of 'EXTERNAL AGENT').",
    )

    object_type: str = Field(
        ..., description="Type of the managing object (e.g. 'EXTERNAL AGENT')."
    )

    object_version: Optional[str] = Field(
        default=None, description="Version of the managing object."
    )

    run_name: str = Field(
        ...,
        description="Unique name of the run. This name should be unique within the object.",
    )

    run_status: Optional[str] = Field(
        default=None,
        description="The status of the run on the entity level. Currently it can only be ACTIVE or CANCELLED.",
    )

    description: Optional[str] = Field(
        default=None, description="A description for the run."
    )

    class RunMetadata(BaseModel):
        labels: Optional[List[str]] = Field(
            default=None,
            description="Text label to group the runs. Take a single label for now",
        )
        llm_judge_name: Optional[str] = Field(
            default=None,
            description="Name of the LLM judge to be used for the run.",
        )
        invocations: Optional[Dict[str, Run.InvocationMetadata]] = Field(
            default=None,
            description="Map of invocation metadata with invocation ID as key.",
        )
        computations: Optional[Dict[str, Run.ComputationMetadata]] = Field(
            default=None,
            description="Map of computation metadata with computation ID as key.",
        )
        metrics: Optional[Dict[str, Run.MetricsMetadata]] = Field(
            default=None,
            description="Map of metrics metadata with metric ID as key.",
        )

    run_metadata: RunMetadata = Field(
        default=...,
        description="Run metadata that maintains states needed for app invocation and metrics computation.",
    )

    class SourceInfo(BaseModel):
        name: str = Field(
            ...,
            description="Name of the source (e.g. name of the table).",
        )
        column_spec: Dict[str, str] = Field(
            default=...,
            description="Column name mapping from reserved dataset fields to column names in user's table.",
        )
        source_type: str = Field(
            default="DATAFRAME",
            description="Type of the source (e.g. 'DATAFRAME').",
        )

    source_info: SourceInfo = Field(
        default=...,
        description="Source information for the run.",
    )

    class CompletionStatusStatus(str, Enum):
        UNKNOWN = "UNKNOWN"
        PARTIALLY_COMPLETED = "PARTIALLY_COMPLETED"
        COMPLETED = "COMPLETED"
        FAILED = "FAILED"

    class CompletionStatus(BaseModel):
        status: Run.CompletionStatusStatus = Field(
            ..., description="The status of the completion."
        )
        record_count: Optional[int] = Field(
            default=None, description="The count of records processed."
        )
        model_config = ConfigDict(json_encoders={Enum: lambda o: o.value})

    class InvocationMetadata(BaseModel):
        input_records_count: Optional[int] = Field(
            default=None,
            description="The number of input records in the dataset.",
        )
        id: Optional[str] = Field(
            default=None,
            description="The unique identifier for the invocation metadata.",
        )
        start_time_ms: Optional[int] = Field(
            default=None,
            description="The start time of the invocation.",
        )
        end_time_ms: Optional[int] = Field(
            default=None,
            description="The end time of the invocation.",
        )
        completion_status: Optional[Run.CompletionStatus] = Field(
            default=None,
            description="The status of the invocation.",
        )

    class ComputationMetadata(BaseModel):
        id: Optional[str] = Field(
            default=None,
            description="Unique id, even if name is repeated.",
        )
        query_id: Optional[str] = Field(
            default=None,
            description="Query id associated with metric computation.",
        )
        start_time_ms: Optional[int] = Field(
            default=None,
            description="Start time of the computation in milliseconds.",
        )
        end_time_ms: Optional[int] = Field(
            default=None,
            description="End time of the computation in milliseconds.",
        )

    class MetricsMetadata(BaseModel):
        id: Optional[str] = Field(
            default=None,
            description="Unique id for the metrics metadata.",
        )
        name: Optional[str] = Field(
            default=None,
            description="Name of the metric.",
        )
        completion_status: Optional[Run.CompletionStatus] = Field(
            default=None,
            description="Completion status of the metric computation.",
        )
        computation_id: Optional[str] = Field(
            default=None,
            description="ID of the computation associated with the metric.",
        )

    def describe(self) -> dict:
        """
        Retrieve the metadata of the Run object.
        """

        run_metadata_df = self.run_dao.get_run(
            run_name=self.run_name,
            object_name=self.object_name,
            object_type=self.object_type,
            object_version=self.object_version,
        )
        if run_metadata_df.empty:
            raise ValueError(f"Run {self.run_name} not found.")

        raw_json = json.loads(
            list(run_metadata_df.to_dict(orient="records")[0].values())[0]
        )

        # remove / hide entity-level run_status field to avoid customer's confusion
        raw_json.pop("run_status", None)
        return raw_json

    def delete(self) -> None:
        """
        Delete the run by its name and object name.
        """
        self.run_dao.delete_run(
            run_name=self.run_name,
            object_name=self.object_name,
            object_type=self.object_type,
            object_version=self.object_version,
        )

    def _can_start_new_invocation(self, current_run_status: RunStatus) -> bool:
        """
        Check if the run is in a state that allows starting a new invocation.
        """
        if self._is_cancelled():
            logger.warning("Cannot start a new invocation for a cancelled run.")
            return False

        return current_run_status in [
            RunStatus.CREATED,
            RunStatus.INVOCATION_PARTIALLY_COMPLETED,
            RunStatus.FAILED,
            RunStatus.UNKNOWN,
        ]

    def _can_start_new_metric_computation(
        self, current_run_status: RunStatus
    ) -> bool:
        """
        Check if the run is in a state that allows starting a new metric computation.
        """
        if self._is_cancelled():
            logger.warning(
                "Cannot start a new metric computation for a cancelled run."
            )
            return False

        if current_run_status == RunStatus.COMPUTATION_IN_PROGRESS:
            logger.warning(
                "Previous computation(s) still in progress. Starting another new metric computation when computation is in progress."
            )

        return current_run_status in [
            RunStatus.INVOCATION_COMPLETED,
            RunStatus.INVOCATION_PARTIALLY_COMPLETED,
            RunStatus.COMPUTATION_IN_PROGRESS,
            RunStatus.COMPLETED,
            RunStatus.PARTIALLY_COMPLETED,
            RunStatus.FAILED,
        ]

    def _is_invocation_started(self, run: Run) -> bool:
        return (
            run.run_metadata.invocations is not None
            and len(run.run_metadata.invocations) > 0
        )

    def _compute_latest_invocation_status(self, run: Run) -> RunStatus:
        latest_invocation = max(
            run.run_metadata.invocations.values(),
            key=lambda inv: (inv.start_time_ms or 0, inv.id or ""),
        )

        if (
            latest_invocation.completion_status
            and latest_invocation.completion_status.status
        ):
            completion_status = latest_invocation.completion_status.status
            if completion_status == Run.CompletionStatusStatus.COMPLETED:
                return RunStatus.INVOCATION_COMPLETED
            elif (
                completion_status
                == Run.CompletionStatusStatus.PARTIALLY_COMPLETED
            ):
                return RunStatus.INVOCATION_PARTIALLY_COMPLETED
            elif completion_status == Run.CompletionStatusStatus.FAILED:
                return RunStatus.FAILED
            else:
                logger.warning(
                    f"Unknown completion status {completion_status} for invocation {latest_invocation.id}"
                )
                return RunStatus.UNKNOWN

        current_ingested_records_count = (
            self.run_dao.read_spans_count_from_event_table(
                object_name=self.object_name,
                run_name=self.run_name,
                span_type="record_root",
            )
        )
        logger.info(
            f"Current ingested records count: {current_ingested_records_count}"
        )

        latest_record_root_timestamp_in_ms = (
            self.run_dao.read_latest_record_root_timestamp_in_ms(
                object_name=self.object_name, run_name=self.run_name
            )
        )

        logger.debug(
            f"Latest record root timestamp in ms: {latest_record_root_timestamp_in_ms}"
        )

        if (
            latest_invocation.input_records_count
            and current_ingested_records_count
            >= latest_invocation.input_records_count
        ):
            # greater than or equal to input records count to account for the edge case where multiple tru recorders are set on the same run
            # happy case, add end time and update status
            self.run_dao.upsert_run_metadata_fields(
                entry_type=SupportedEntryType.INVOCATIONS.value,
                entry_id=latest_invocation.id,
                input_records_count=latest_invocation.input_records_count,
                start_time_ms=latest_invocation.start_time_ms,
                end_time_ms=self._get_current_time_in_ms(),
                completion_status=Run.CompletionStatus(
                    status=Run.CompletionStatusStatus.COMPLETED,
                    record_count=current_ingested_records_count,
                ).model_dump(),
                run_name=self.run_name,
                object_name=self.object_name,
                object_type=self.object_type,
                object_version=self.object_version,
            )

            return RunStatus.INVOCATION_COMPLETED

        elif (
            latest_invocation.start_time_ms
            and latest_record_root_timestamp_in_ms
            and time.time() * 1000 - latest_record_root_timestamp_in_ms
            > EXPECTED_TELEMETRY_LATENCY_IN_MS
        ):
            # inconclusive case, timeout reached and add end time and update completion status in DPO
            logger.warning("Invocation timed out.")
            self.run_dao.upsert_run_metadata_fields(
                entry_type=SupportedEntryType.INVOCATIONS.value,
                entry_id=latest_invocation.id,
                input_records_count=latest_invocation.input_records_count,
                start_time_ms=latest_invocation.start_time_ms,
                end_time_ms=self._get_current_time_in_ms(),
                completion_status=Run.CompletionStatus(
                    status=Run.CompletionStatusStatus.PARTIALLY_COMPLETED,
                    record_count=current_ingested_records_count,
                ).model_dump(),
                run_name=self.run_name,
                object_name=self.object_name,
                object_type=self.object_type,
                object_version=self.object_version,
            )

            return RunStatus.INVOCATION_PARTIALLY_COMPLETED

        else:
            return (
                RunStatus.INVOCATION_IN_PROGRESS
                if latest_invocation.end_time_ms == 0
                else RunStatus.UNKNOWN
            )

    def _metrics_computation_started(self, run: Run) -> bool:
        return (
            run.run_metadata.metrics is not None
            and len(run.run_metadata.metrics) > 0
        )

    def _resolve_overall_metrics_status(
        self, all_metrics, invocation_completion_status
    ):
        """
        Given the list of all metrics, resolve and return the overall metrics status.
        """
        if all(
            metric.completion_status
            and metric.completion_status.status
            == Run.CompletionStatusStatus.COMPLETED
            for metric in all_metrics
        ):
            return (
                RunStatus.COMPLETED
                if invocation_completion_status
                == Run.CompletionStatusStatus.COMPLETED
                else RunStatus.PARTIALLY_COMPLETED
            )
        elif all(
            metric.completion_status
            and metric.completion_status.status
            == Run.CompletionStatusStatus.FAILED
            for metric in all_metrics
        ):
            return RunStatus.FAILED
        elif all(
            metric.completion_status
            and metric.completion_status.status
            in [
                Run.CompletionStatusStatus.COMPLETED,
                Run.CompletionStatusStatus.FAILED,
            ]
            for metric in all_metrics
        ):
            return RunStatus.PARTIALLY_COMPLETED
        else:
            logger.warning(
                "Cannot determine run status. Metrics: %s",
                [metric.name for metric in all_metrics],
            )
            return RunStatus.UNKNOWN

    def _compute_overall_computations_status(self, run: Run) -> RunStatus:
        all_existing_metrics = run.run_metadata.metrics.values()

        latest_invocation = max(
            run.run_metadata.invocations.values(),
            key=lambda inv: (inv.start_time_ms or 0, inv.id or ""),
        )
        invocation_completion_status = (
            latest_invocation.completion_status.status
        )

        # check all metrics and see if their completion status are set or not
        metrics_status_not_set = [
            metric
            for metric in all_existing_metrics
            if not metric.completion_status
            or not metric.completion_status.status
        ]

        if len(metrics_status_not_set) == 0:
            # early return cases as status of all metrics are set
            logger.info("All metrics statuses are set.")
            return self._resolve_overall_metrics_status(
                all_existing_metrics, invocation_completion_status
            )
        else:
            logger.info(
                f"Metrics status not set for: {[metric.name for metric in metrics_status_not_set]}. Checking sproc query status via query history"
            )
            # multiple metrics can be associated with the same computation
            computation_id_to_metrics = defaultdict(list)
            for metric in metrics_status_not_set:
                computation_id_to_metrics[metric.computation_id].append(metric)

            all_computations = run.run_metadata.computations.values()
            # Check the status of sproc query with metrics that are not complete,
            # update the ones that are done.

            some_computation_in_progress = False
            for computation in all_computations:
                if computation.id in computation_id_to_metrics:
                    query_id = computation.query_id
                    query_status = (
                        self.run_dao.fetch_query_execution_status_by_id(
                            query_start_time_ms=computation.start_time_ms,
                            query_id=query_id,
                        )
                    )
                    if query_status == "IN_PROGRESS":
                        logger.info(
                            f"Computation {computation.id} is still running or being queued."
                        )
                        some_computation_in_progress = True

                    elif query_status == "FAILED" or query_status == "SUCCESS":
                        logger.warning(
                            f"Computation query_id: {query_id} finished with status: {query_status}. Updating run metadata."
                        )

                        self.run_dao.upsert_run_metadata_fields(
                            entry_type=SupportedEntryType.COMPUTATIONS.value,
                            entry_id=computation.id,
                            query_id=query_id,
                            start_time_ms=computation.start_time_ms,
                            end_time_ms=self._get_current_time_in_ms(),
                            run_name=self.run_name,
                            object_name=self.object_name,
                            object_type=self.object_type,
                            object_version=self.object_version,
                        )
                        metrics_in_computation = computation_id_to_metrics[
                            computation.id
                        ]
                        result_rows = self.run_dao.fetch_computation_job_results_by_query_id(
                            query_id
                        )

                        metric_name_to_status = {
                            row["METRIC"]: row["STATUS"]
                            for _, row in result_rows.iterrows()
                        }

                        metric_name_to_computed_records_count = {
                            row["METRIC"]: int(
                                row["MESSAGE"].split(" ")[
                                    1
                                ]  # TODO unbrittel this - directly read the field when available
                            )
                            for _, row in result_rows.iterrows()
                        }

                        for metric in metrics_in_computation:
                            if (
                                metric.name in metric_name_to_status
                                and metric.name
                                in metric_name_to_computed_records_count
                            ):
                                logger.info(
                                    f"Updating metric {metric.name} metadata."
                                )
                                self.run_dao.upsert_run_metadata_fields(
                                    entry_type=SupportedEntryType.METRICS.value,
                                    entry_id=metric.id,
                                    computation_id=computation.id,
                                    name=metric.name,
                                    completion_status=Run.CompletionStatus(
                                        status=Run.CompletionStatusStatus.COMPLETED
                                        if metric_name_to_status[metric.name]
                                        == "SUCCESS"
                                        else Run.CompletionStatusStatus.FAILED,
                                        record_count=metric_name_to_computed_records_count[
                                            metric.name
                                        ],  # TODO: read from event table if possible
                                    ).model_dump(),
                                    run_name=self.run_name,
                                    object_name=self.object_name,
                                    object_type=self.object_type,
                                    object_version=self.object_version,
                                )

            if some_computation_in_progress:
                return RunStatus.COMPUTATION_IN_PROGRESS
            else:
                logger.info("All computations concluded.")
                return self._resolve_overall_metrics_status(
                    all_existing_metrics, invocation_completion_status
                )

    def start(self, input_df: Optional[pd.DataFrame] = None):
        """
        Start the run by invoking the main method of the user's app with the input data

        Args:
            input_df (Optional[pd.DataFrame], optional): user provided input dataframe.
        """
        current_status = self.get_status()
        logger.info(f"Current run status: {current_status}")
        if not self._can_start_new_invocation(current_status):
            return f"Cannot start a new invocation when in run status: {current_status}. Valid statuses are: {RunStatus.CREATED}, {RunStatus.INVOCATION_PARTIALLY_COMPLETED}, or {RunStatus.FAILED}."

        if input_df is None:
            logger.info(
                "No input dataframe provided. Fetching input data from source."
            )
            rows = self.run_dao.session.sql(
                f"SELECT * FROM {self.source_info.name}"
            ).collect()
            input_df = pd.DataFrame([row.as_dict() for row in rows])

        dataset_spec = self.source_info.column_spec

        # Preprocess the dataset_spec to create mappings for input columns
        # and map the inputs for reserved fields only once, before the iteration over rows.

        reserved_field_column_mapping = {}

        # Process dataset column spec to handle subscripting logic for input columns
        for reserved_field, user_column in dataset_spec.items():
            reserved_field_column_mapping[reserved_field] = user_column

        input_records_count = len(input_df)

        invocation_metadata_id = self.run_dao._compute_invocation_metadata_id(
            dataset_name=self.source_info.name,
            input_records_count=input_records_count,
        )
        start_time_ms = self._get_current_time_in_ms()

        logger.info(
            f"Creating or updating invocation metadata with {input_records_count} records from input."
        )

        self.run_dao.upsert_run_metadata_fields(
            entry_type=SupportedEntryType.INVOCATIONS.value,
            entry_id=invocation_metadata_id,
            start_time_ms=start_time_ms,
            end_time_ms=0,  # required field
            run_name=self.run_name,
            input_records_count=input_records_count,
            object_name=self.object_name,
            object_type=self.object_type,
            object_version=self.object_version,
        )

        # user app invocation - will block until the app completes
        try:
            for i, row in input_df.iterrows():
                main_method_args = []

                # Call the instrumented main method with the arguments
                # TODO (dhuang) better way to check span attributes, also is this all we need to support?
                input_id = (
                    row[dataset_spec["input_id"]]
                    if "input_id" in dataset_spec
                    else None
                )
                input_col = None
                if input_id is None:
                    if "input" in dataset_spec:
                        input_col = dataset_spec["input"]
                    elif "record_root.input" in dataset_spec:
                        input_col = dataset_spec["record_root.input"]
                    if input_col:
                        input_id = obj_id_of_obj(row[input_col])
                        main_method_args.append(row[input_col])

                ground_truth_output = row.get(
                    dataset_spec.get("ground_truth_output")
                    or dataset_spec.get("record_root.ground_truth_output")
                )

                self.app.instrumented_invoke_main_method(
                    run_name=self.run_name,
                    input_id=input_id,
                    ground_truth_output=ground_truth_output,
                    main_method_args=tuple(
                        main_method_args
                    ),  # Ensure correct order
                    main_method_kwargs=None,  # don't take any kwargs for now so we don't break TruChain / TruLlama where input argument name cannot be defined by users.
                )
        except Exception as e:
            logger.exception(
                f"Error encountered during invoking app main method: {e}."
            )

            self.run_dao.upsert_run_metadata_fields(
                entry_type=SupportedEntryType.INVOCATIONS.value,
                entry_id=invocation_metadata_id,
                start_time_ms=start_time_ms,
                input_records_count=input_records_count,
                end_time_ms=self._get_current_time_in_ms(),
                completion_status=Run.CompletionStatus(
                    status=Run.CompletionStatusStatus.FAILED,
                ).model_dump(),
                run_name=self.run_name,
                object_name=self.object_name,
                object_type=self.object_type,
                object_version=self.object_version,
            )

            raise

        self.tru_session.force_flush()
        logger.info("Run started, invocation done and ingestion in process.")

    def _get_current_time_in_ms(self) -> int:
        return int(round(time.time() * 1000))

    def get_status(self) -> RunStatus:
        run_metadata_df = self.run_dao.get_run(
            run_name=self.run_name,
            object_name=self.object_name,
            object_type=self.object_type,
            object_version=self.object_version,
        )

        run = Run.from_metadata_df(
            run_metadata_df,
            {
                "app": self,
                "main_method_name": self.main_method_name,
                "run_dao": self.run_dao,
                "tru_session": self.tru_session,
            },
        )

        if run.run_status == "CANCELLED":
            return RunStatus.CANCELLED

        if not self._is_invocation_started(run):
            logger.info("Run is created, no invocation nor computation yet.")
            return RunStatus.CREATED
        elif self._metrics_computation_started(run):
            logger.info(
                "Run is created, invocation done, and some or all metrics computed."
            )

            return self._compute_overall_computations_status(run)

        else:
            logger.info("Run is created, invocation started.")

            return self._compute_latest_invocation_status(run)

    def _should_skip_computation(self, metric_name: str, run: Run) -> bool:
        if run.run_metadata.metrics is None:
            return False

        statuses = []  # will store statuses for all matching metric entries
        for metric_metadata in run.run_metadata.metrics.values():
            if metric_metadata.name == metric_name:
                # If completion_status is not set, we treat it as "in progress"
                if metric_metadata.completion_status is None:
                    statuses.append("IN_PROGRESS")
                else:
                    statuses.append(metric_metadata.completion_status.status)

        # If no matching metric entries found, don't skip.
        if not statuses:
            return False

        if any(s == Run.CompletionStatusStatus.COMPLETED for s in statuses):
            logger.info(
                f"Metric {metric_name} already computed successfully (one entry COMPLETED); skipping computation."
            )
            return True

        # If any metric is in progress (i.e. no completion status), we skip because it's still computing.
        if any(s == "IN_PROGRESS" for s in statuses):
            logger.info(
                f"Metric {metric_name} is in progress (at least one entry not complete); skipping computation."
            )
            return True

        # If all matching metrics are FAILED, then allow re-computation.
        if all(s == Run.CompletionStatusStatus.FAILED for s in statuses):
            logger.info(
                f"All metric entries for {metric_name} have FAILED; allowing re-computation."
            )
            return False

        logger.warning(
            "Unknown state for metric computation; skipping computation."
        )
        return True

    def compute_metrics(self, metrics: List[str]) -> str:
        run_status = self.get_status()

        logger.info(f"Current run status: {run_status}")
        if not self._can_start_new_metric_computation(run_status):
            return f"""Cannot start a new metric computation when in run status: {run_status}. Valid statuses are: {RunStatus.INVOCATION_COMPLETED}, {RunStatus.INVOCATION_PARTIALLY_COMPLETED},
        {RunStatus.COMPUTATION_IN_PROGRESS}, {RunStatus.COMPLETED}, {RunStatus.PARTIALLY_COMPLETED}, {RunStatus.FAILED}."""

        run_metadata_df = self.run_dao.get_run(
            run_name=self.run_name,
            object_name=self.object_name,
            object_type=self.object_type,
            object_version=self.object_version,
        )

        run = Run.from_metadata_df(
            run_metadata_df,
            {
                "app": self,
                "main_method_name": self.main_method_name,
                "run_dao": self.run_dao,
                "tru_session": self.tru_session,
            },
        )

        computation_metadata_id = str(uuid.uuid4())

        for metric_name in metrics:
            if not self._should_skip_computation(metric_name, run):
                logger.info(
                    f"Adding metric: {metric_name} to run metadata for computation."
                )
                # add placeholder entries to metrics field in run metadata
                metric_metadata_id = str(uuid.uuid4())
                self.run_dao.upsert_run_metadata_fields(
                    entry_type=SupportedEntryType.METRICS.value,
                    entry_id=metric_metadata_id,
                    computation_id=computation_metadata_id,
                    name=metric_name,
                    completion_status=None,  # starting w/ null, will be updated after the computation
                    run_name=self.run_name,
                    object_name=self.object_name,
                    object_type=self.object_type,
                    object_version=self.object_version,
                )

        computation_start_time_ms = self._get_current_time_in_ms()

        async_job = self.run_dao.call_compute_metrics_query(
            metrics=metrics,
            object_name=self.object_name,
            object_version=self.object_version,
            object_type=self.object_type,
            run_name=self.run_name,
        )

        query_id = async_job.query_id

        logger.info(f"Query id for metrics computation: {query_id}")
        self.run_dao.upsert_run_metadata_fields(
            entry_type=SupportedEntryType.COMPUTATIONS.value,
            entry_id=computation_metadata_id,
            query_id=query_id,
            start_time_ms=computation_start_time_ms,
            end_time_ms=0,
            run_name=self.run_name,
            object_name=self.object_name,
            object_type=self.object_type,
            object_version=self.object_version,
        )

        logger.info("Metrics computation job started")
        return "Metrics computation in progress."

    def _is_cancelled(self) -> bool:
        return self.get_status() == RunStatus.CANCELLED

    def cancel(self):
        if self._is_cancelled():
            logger.warning(f"Run {self.run_name} is already cancelled.")
            return

        update_fields = {"run_status": "CANCELLED"}

        self.run_dao.upsert_run_metadata_fields(
            run_name=self.run_name,
            object_name=self.object_name,
            object_type=self.object_type,
            object_version=self.object_version,
            **update_fields,
        )

        logger.info(f"Run {self.run_name} cancelled.")

    def update(
        self, description: Optional[str] = None, label: Optional[str] = None
    ):
        """
        Only description and label are allowed to be updated at the moment.
        """
        update_fields = {}
        if description is not None:
            logger.info(f"Updating run description to {description}")
            update_fields["description"] = description
        if label is not None:
            logger.info(f"Updating run label to {label}")
            update_fields["labels"] = [label]

        if update_fields:
            self.run_dao.upsert_run_metadata_fields(
                run_name=self.run_name,
                object_name=self.object_name,
                object_type=self.object_type,
                object_version=self.object_version,
                **update_fields,
            )

    @classmethod
    def from_metadata_df(
        cls, metadata_df: pd.DataFrame, extra: Dict[str, Any]
    ) -> Run:
        """
        Create a Run instance from a metadata DataFrame returned by the DAO,
        and enrich it with additional fields (which are not persisted on the server).

        Args:
            metadata_df: A pandas DataFrame containing run metadata.
                We assume the first row contains a JSON string in its first cell.
            extra: A dictionary of extra fields to add, such as:
                {
                    "app": <app instance>,
                    "main_method_name": <method name>,
                    "run_dao": <dao instance>,
                    "object_name": <object name>,
                    "object_type": <object type>
                }

        Returns:
            A validated Run instance.
        """
        if metadata_df.empty:
            raise ValueError("No run metadata found.")

        # Assume the first cell of the first row contains the JSON string.

        metadata_str = metadata_df.iloc[0].values[0]
        try:
            metadata = json.loads(metadata_str)
        except json.JSONDecodeError:
            raise ValueError(
                "The first cell of the first row does not contain a valid JSON string."
            )

        metadata.update(extra)

        return cls.model_validate(metadata)



"""Import utilities for required and optional imports.

Utilities for importing Python modules and optional importing.
"""

import builtins
from dataclasses import dataclass
from importlib import metadata
import inspect
from inspect import cleandoc
import logging
from pathlib import Path
from pprint import PrettyPrinter
import sys
from typing import (
    Any,
    Dict,
    Iterable,
    Optional,
    Sequence,
    Type,
    Union,
)

import importlib_resources
from packaging import requirements
from packaging import version
import pkg_resources

logger = logging.getLogger(__name__)
pp = PrettyPrinter()


def safe_importlib_package_name(package_name: str) -> str:
    """Convert a package name that may have periods in it to one that uses
    hyphens for periods but only if the Python version is old."""

    return (
        package_name
        if sys.version_info >= (3, 10)
        else package_name.replace(".", "-")
    )


def _requirements_of_trulens_core_file(
    path: str,
) -> Dict[str, requirements.Requirement]:
    """Get a dictionary of package names to requirements from a requirements
    file in trulens.core."""
    _trulens_resources = importlib_resources.files("trulens.core")
    with importlib_resources.as_file(_trulens_resources / path) as _path:
        with open(_path) as fh:
            reqs = pkg_resources.parse_requirements(fh)

            mapping = {}

            for req in reqs:
                mapping[req.name] = req

            return mapping


def static_resource(namespace: str, filepath: Union[Path, str]) -> Path:
    """Get the path to a static resource file in the trulens package.

    By static here we mean something that exists in the filesystem already and
    not in some temporary folder. We use the `importlib.resources` context
    managers to get this but if the resource is temporary, the result might not
    exist by the time we return or is not expected to survive long.
    """

    if not isinstance(filepath, Path):
        filepath = Path(filepath)

    if sys.version_info >= (3, 9):
        # This does not exist in 3.8
        from importlib.abc import Traversable

        _trulens_resources: Traversable = importlib_resources.files(
            f"trulens.{namespace}"
        )
        with importlib_resources.as_file(
            _trulens_resources / filepath
        ) as _path:
            return _path
    else:
        # This is deprecated starting 3.11
        parts = filepath.parts
        with importlib_resources.path("trulens", parts[0]) as _path:
            # NOTE: resources.path does not allow the resource to include folders.
            for part in parts[1:]:
                _path = _path / part
            return _path


def parse_version(version_string: str) -> version.Version:
    """Parse the version string into a packaging version object."""

    return version.parse(version_string)


def get_package_version(name: str) -> Optional[version.Version]:
    """Get the version of a package by its name.

    Returns None if given package is not installed.
    """

    try:
        return parse_version(metadata.version(name))

    except metadata.PackageNotFoundError:
        return None


def is_package_installed(name: str) -> bool:
    """Check if a package is installed."""

    return get_package_version(name) is not None


MESSAGE_DEBUG_OPTIONAL_PACKAGE_NOT_FOUND = """Optional package %s is not installed. Related optional functionality will not
be available.
"""

MESSAGE_ERROR_REQUIRED_PACKAGE_NOT_FOUND = """Required package {req.name} is not installed. Please install it with pip:

    ```bash
    pip install '{req}'
    ```

If your distribution is in a bad place beyond this package, you may need to
reinstall trulens so that all of the dependencies get installed:

    ```bash
    pip uninstall -y trulens
    pip install trulens
    ```
"""

MESSAGE_FRAGMENT_VERSION_MISMATCH = """Package {req.name} is installed but has a version conflict:
    Requirement: {req}
    Installed: {dist.version}
"""

MESSAGE_FRAGMENT_VERSION_MISMATCH_OPTIONAL = """This package is optional for trulens so this may not be a problem but if
you need to use the related optional features and find there are errors, you
will need to resolve the conflict:
"""

MESSAGE_FRAGMENT_VERSION_MISMATCH_REQUIRED = """This package is required for trulens. Please resolve the conflict by
installing a compatible version with:
"""

MESSAGE_FRAGMENT_VERSION_MISMATCH_PIP = """
    ```bash
    pip install '{req}'
    ```

If you are running trulens in a notebook, you may need to restart the
kernel after resolving the conflict. If your distribution is in a bad place
beyond this package, you may need to reinstall trulens so that all of the
dependencies get installed and hopefully corrected:

    ```bash
    pip uninstall -y trulens
    pip install trulens
    ```
"""

required_packages: Dict[str, requirements.Requirement] = (
    _requirements_of_trulens_core_file("utils/requirements.txt")
)
"""Mapping of required package names to the requirement object with info
about that requirement including version constraints."""

optional_packages: Dict[str, requirements.Requirement] = (
    _requirements_of_trulens_core_file("utils/requirements.optional.txt")
)
"""Mapping of optional package names to the requirement object with info
about that requirement including version constraints."""

all_packages: Dict[str, requirements.Requirement] = {
    **required_packages,
    **optional_packages,
}
"""Mapping of optional and required package names to the requirement object
with info about that requirement including version constraints."""


class VersionConflict(Exception):
    """Exception to raise when a version conflict is found in a required package."""


def check_imports(ignore_version_mismatch: bool = False):
    """Check required and optional package versions.
    Args:
        ignore_version_mismatch: If set, will not raise an error if a
            version mismatch is found in a required package. Regardless of
            this setting, mismatch in an optional package is a warning.
    Raises:
        VersionConflict: If a version mismatch is found in a required package
            and `ignore_version_mismatch` is not set.
    """

    for n, req in all_packages.items():
        is_optional = n in optional_packages

        try:
            dist = metadata.distribution(req.name)

        except metadata.PackageNotFoundError as e:
            if is_optional:
                logger.debug(MESSAGE_DEBUG_OPTIONAL_PACKAGE_NOT_FOUND, req.name)

            else:
                raise ModuleNotFoundError(
                    MESSAGE_ERROR_REQUIRED_PACKAGE_NOT_FOUND.format(req=req)
                ) from e

        if dist.version not in req.specifier:
            message = MESSAGE_FRAGMENT_VERSION_MISMATCH.format(
                req=req, dist=dist
            )

            if is_optional:
                message += MESSAGE_FRAGMENT_VERSION_MISMATCH_OPTIONAL.format(
                    req=req
                )

            else:
                message += MESSAGE_FRAGMENT_VERSION_MISMATCH_REQUIRED.format(
                    req=req
                )

            message += MESSAGE_FRAGMENT_VERSION_MISMATCH_PIP.format(req=req)

            if (not is_optional) and (not ignore_version_mismatch):
                raise VersionConflict(message)

            logger.debug(message)


def pin_spec(r: requirements.Requirement) -> requirements.Requirement:
    """
    Pin the requirement to the version assuming it is lower bounded by a
    version.
    """

    spec = str(r)
    if ">=" not in spec:
        raise ValueError(f"Requirement {spec} is not lower-bounded.")

    spec = spec.replace(">=", "==")
    return requirements.Requirement(spec)


@dataclass
class ImportErrorMessages:
    """Container for messages to show when an optional package is not found or
    has some other import error."""

    module_not_found: str
    """Message to show or raise when a package is not found."""

    import_error: str
    """Message to show or raise when a package may be installed but some import
    error occurred trying to import it or something from it."""


def format_import_errors(
    packages: Union[str, Sequence[str]],
    purpose: Optional[str] = None,
    throw: Union[bool, Exception] = False,
) -> ImportErrorMessages:
    """Format two messages for missing optional package or bad import from an
    optional package.

    Throws an `ImportError` with the formatted message if `throw` flag is set.
    If `throw` is already an exception, throws that instead after printing the
    message.
    """

    if purpose is None:
        purpose = f"using {packages}"

    if isinstance(packages, str):
        packages = [packages]

    requirements = []
    requirements_pinned = []

    for pkg in packages:
        if pkg in all_packages:
            requirements.append(str(all_packages[pkg]))
            requirements_pinned.append(str(pin_spec(all_packages[pkg])))
        else:
            logger.warning("Package %s not present in requirements.", pkg)
            requirements.append(pkg)

    packs = ",".join(packages)
    pack_s = "package" if len(packages) == 1 else "packages"
    is_are = "is" if len(packages) == 1 else "are"
    it_them = "it" if len(packages) == 1 else "them"
    this_these = "this" if len(packages) == 1 else "these"

    msg = cleandoc(f"""
These {pack_s} {is_are} required for {purpose}:

    {",".join(packages)}

You should be able to install {it_them} with pip:

    ```bash
    pip install {" ".join(map(lambda a: f'"{a}"', requirements))}
    ```
""")

    msg_pinned = cleandoc(f"""
There was a problem using these packages:

    {packs}

They are required for {purpose}. Try installing {this_these} exact {pack_s} with pip:

    ```bash
    pip install {" ".join(map(lambda a: f'"{a}"', requirements_pinned))}
    ```
""")

    if isinstance(throw, Exception):
        print(msg)
        raise throw

    elif isinstance(throw, bool):
        if throw:
            raise ImportError(msg)

    return ImportErrorMessages(module_not_found=msg, import_error=msg_pinned)


def is_dummy(obj: Any) -> bool:
    """Check if the given object is an instance of `Dummy`.

    This is necessary as `isisintance` and `issubclass` checks might fail if the
    ones defined in `Dummy` get used; they always return `False` by design.
    """

    return obj.__class__.__name__ == "Dummy" and obj.__module__ in [
        "trulens.core.utils.imports",
        "trulens_eval.utils.imports",
    ]


# Try to pretend to be a type as well as an instance.
class Dummy(type):
    """Class to pretend to be a module or some other imported object.

    Will raise an error if accessed in some dynamic way. Accesses that are
    "static-ish" will try not to raise the exception so things like defining
    subclasses of a missing class should not raise exception. Dynamic uses are
    things like calls, use in expressions. Looking up an attribute is static-ish
    so we don't throw the error at that point but instead make more dummies.


    Warning:
        While dummies can be used as types, they return false to all `isinstance`
        and `issubclass` checks. Further, the use of a dummy in subclassing
        produces unreliable results with some of the debugging information such
        as `original_exception` may be inaccessible.
    """

    def __str__(self) -> str:
        ret = f"Dummy({self.name}"

        if self.original_exception is not None:
            ret += f", original_exception={self.original_exception}"

        ret += ")"

        return ret

    def __repr__(self) -> str:
        return str(self)

    def __new__(cls, name, *args, **kwargs):
        if (
            len(args) >= 2
            and isinstance(args[1], dict)
            and "__classcell__" in args[1]
        ):
            # Used as type, for subclassing for example.

            return type.__new__(cls, name, args[0], args[1])
        else:
            return type.__new__(cls, name, (cls,), kwargs)

    def __init__(self, name: str, *args, **kwargs):
        if len(args) >= 2 and isinstance(args[1], dict):
            # Used as type, in subclassing for example.

            src = args[0][0]

            message = src.message
            importer = src.importer

            original_exception = src.original_exception
            exception_class = src.exception_class

        else:
            message: str = kwargs.get("message", None)
            exception_class: Type[Exception] = kwargs.get(
                "exception_class", ModuleNotFoundError
            )
            importer = kwargs.get("importer", None)
            original_exception: Optional[Exception] = kwargs.get(
                "original_exception", None
            )

        # For easier checking that an instance is Dummy. If may be hard to check
        # for dummy if comparing against something that does not come with its
        # own __isinstance__ as the one we include here always returns False.
        self.__dummy__ = True

        self.name = name
        self.message = message
        self.importer = importer
        self.exception_class = exception_class

        if original_exception is not None:
            # Intentionally not including traceback as it includes frames which include references to
            # objects we don't want to reference:
            original_exception = original_exception.with_traceback(None)

        self.original_exception: Optional[Exception] = original_exception

    def __call__(self, *args, **kwargs):
        raise self.exception_class(self.message) from self.original_exception

    def __instancecheck__(self, __instance: Any) -> bool:
        """Nothing is an instance of this dummy.

        Warning:
            This is to make sure that if something optional gets imported as a
            dummy and is a class to be instrumented, it will not automatically make
            the instrumentation class check succeed on all objects.
        """
        return False

    def __subclasscheck__(self, __subclass: type) -> bool:
        """Nothing is a subclass of this dummy."""

        return False

    def _wasused(self, *args, **kwargs):
        raise self.exception_class(self.message) from self.original_exception

    # If someone tries to use dummy in an expression, raise our usage exception:
    __add__ = _wasused
    __sub__ = _wasused
    __mul__ = _wasused
    __truediv__ = _wasused
    __floordiv__ = _wasused
    __mod__ = _wasused
    __divmod__ = _wasused
    __pow__ = _wasused
    __lshift__ = _wasused
    __rshift__ = _wasused
    __and__ = _wasused
    __xor__ = _wasused
    __or__ = _wasused
    __radd__ = _wasused
    __rsub__ = _wasused

    def __getattr__(self, name):
        # If in OptionalImport context, create a new dummy for the requested attribute. Otherwise
        # raise error.

        if name == "__dummy__":
            return True

        # Pretend to be object for generic attributes.
        if hasattr(object, name):
            return getattr(object, name)

        # Prevent pydantic inspecting this object as if it were a type from triggering the exception
        # message below.
        if name in [
            "__pydantic_generic_metadata__",
            "__get_pydantic_core_schema__",
            "__get_validators__",
            "__get_pydantic_json_schema__",
            "__modify_schema__",
            "__origin__",
            "__dataclass_fields__",
        ]:
            raise AttributeError()

        # If we are still in an optional import block, continue making dummies
        # inside this dummy.
        if self.importer is not None and (
            self.importer.importing and not self.importer.fail
        ):
            return Dummy(
                name=self.name + "." + name,
                message=self.message,
                importer=self.importer,
                original_exception=self.original_exception,
                exception_class=ModuleNotFoundError,
            )

        # If we are no longer in optional imports context or context said to
        # fail anyway, raise the exception with the optional package message.

        raise self.exception_class(self.message)


class OptionalImports:
    """Helper context manager for doing multiple imports from an optional
    modules

    Example:
        ```python
            messages = ImportErrorMessages(
                module_not_found="install llama_index first",
                import_error="install llama_index==0.1.0"
            )
            with OptionalImports(messages=messages):
                import llama_index
                from llama_index import query_engine
        ```

        The above Python block will not raise any errors but once anything else
        about llama_index or query_engine gets accessed, an error is raised with the
        specified message (unless llama_index is installed of course).
    """

    def assert_installed(self, mods: Union[Any, Iterable[Any]]):
        """Check that the given modules `mods` are not dummies. If any is, show the
        optional requirement message.

        Returns self for chaining convenience.
        """
        if not isinstance(mods, Iterable):
            mods = [mods]

        if any(isinstance(mod, Dummy) for mod in mods):
            raise ModuleNotFoundError(self.messages.module_not_found)

        return self

    def __init__(self, messages: ImportErrorMessages, fail: bool = False):
        """
        Create an optional imports context manager class. Will keep module not
        found or import errors quiet inside context unless fail is True.
        """

        self.messages = messages
        self.importing = False
        self.fail = fail
        self.imp = builtins.__import__

    def __import__(
        self,
        name,
        globals: Optional[Dict[str, Any]] = None,
        locals: Optional[Dict[str, Any]] = None,
        fromlist: Sequence[str] = (),
        level=0,
    ):
        # Check if this import call is coming from an import in trulens as
        # otherwise we don't want to intercept the error as some modules rely on
        # import failures for various things. HACK012: we have to step back a
        # frame or two here to check where the original import came from. We
        # skip any frames that refer to our overwritten __import__. We have to
        # step back multiple times if we (accidentally) nest our OptionalImport
        # context manager.
        frame = inspect.currentframe().f_back
        while frame.f_code == self.__import__.__code__:
            frame = frame.f_back

        module_name = frame.f_globals.get("__name__", "")

        if fromlist is None:
            fromlist = ()

        if not module_name.startswith("trulens"):
            # If we are not importing in trulens, ignore this mechanism and use
            # the builtin import. We don't want to override behaviour of
            # importing for other packages as they might rely on import errors
            # to drive their own logic.
            return self.imp(name, globals, locals, fromlist, level)

        try:
            # First check whether the module itself is importable.
            mod = self.imp(name, globals, locals, fromlist, level)

        except ModuleNotFoundError as e:
            if self.fail:
                raise e

            return Dummy(
                name=name,
                message=self.messages.module_not_found,
                importer=self,
                original_exception=e,
                exception_class=ModuleNotFoundError,
            )

        except ImportError as e:
            if self.fail:
                raise e

            return Dummy(
                name=name,
                message=self.messages.import_error,
                exception_class=ImportError,
                importer=self,
                original_exception=e,
            )

        # By this point the module was imported successfully. Now we need to
        # check if we need to make dummies for the fromlist.

        for i in fromlist:
            if i == "*":
                # Don't have a good way of verifying that * imports are all
                # available. We should not be using those in optional imports.
                logger.warning(
                    "Optional import of * from %s is not supported.",
                    name,
                )
                continue

            try:
                val = getattr(mod, i)

            except AttributeError as e:
                if self.fail:
                    raise e

                val = Dummy(
                    name=i,
                    message=self.messages.module_not_found,
                    importer=self,
                    original_exception=e,
                    exception_class=AttributeError,
                )

            except ImportError as e:
                if self.fail:
                    raise e

                val = Dummy(
                    name=i,
                    message=self.messages.import_error,
                    importer=self,
                    original_exception=e,
                    exception_class=ImportError,
                )

            mod.__dict__[i] = val

            if globals is not None:
                globals[i] = val

            if locals is not None:
                locals[i] = val

        return mod

    def __enter__(self):
        """Handle entering the WithOptionalImports context block.

        We override the builtins.__import__ function to catch any import errors.
        """

        # TODO: Better handle nested contexts. For now we don't override import
        # and just let the already-overridden one do its thing.

        if "trulens" in str(builtins.__import__):
            logger.debug(
                "Nested optional imports context used. This context will be ignored."
            )
        else:
            builtins.__import__ = self.__import__

        self.importing = True

        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Handle exiting from the WithOptionalImports context block.

        We should not get any exceptions here if dummies were produced by the
        overwritten __import__ but if an import of a module that exists failed
        becomes some component of that module did not, we will not be able to
        catch it to produce dummy and have to process the exception here in
        which case we add our informative message to the exception and re-raise
        it.
        """
        self.importing = False
        builtins.__import__ = self.imp

        if exc_value is None:
            return

        if isinstance(exc_value, ModuleNotFoundError):
            if exc_value.msg.startswith(self.messages.module_not_found):
                # Don't add more to the message if it already includes our instructions.
                raise exc_value

            raise ModuleNotFoundError(
                self.messages.module_not_found
            ) from exc_value

        elif isinstance(exc_value, ImportError):
            if exc_value.msg.startswith(self.messages.import_error):
                # Don't add more to the message if it already includes our instructions.
                raise exc_value

            raise ImportError(self.messages.import_error) from exc_value

        # Exception will be propagated unless we return True so we don't return it.



import logging
from typing import ClassVar, Optional

from trulens.core.feedback import endpoint as core_endpoint
from trulens.core.utils import pyschema as pyschema_utils
from trulens.core.utils import serial as serial_utils

logger = logging.getLogger(__name__)


class Provider(pyschema_utils.WithClassInfo, serial_utils.SerialModel):
    """Base Provider class.

    TruLens makes use of *Feedback Providers* to generate evaluations of
    large language model applications. These providers act as an access point
    to different models, most commonly classification models and large language models.

    These models are then used to generate feedback on application outputs or intermediate
    results.

    `Provider` is the base class for all feedback providers. It is an abstract
    class and should not be instantiated directly. Rather, it should be subclassed
    and the subclass should implement the methods defined in this class.

    There are many feedback providers available in TruLens that grant access to a wide range
    of proprietary and open-source models.

    Providers for classification and other non-LLM models should directly subclass `Provider`.
    The feedback functions available for these providers are tied to specific providers, as they
    rely on provider-specific endpoints to models that are tuned to a particular task.

    For example, the HuggingFace feedback provider provides access to a number of classification models
    for specific tasks, such as language detection. These models are than utilized by a feedback function
    to generate an evaluation score.

    Example:
        ```python
        from trulens.providers.huggingface import Huggingface
        huggingface_provider = Huggingface()
        huggingface_provider.language_match(prompt, response)
        ```

    Providers for LLM models should subclass `trulens.feedback.llm_provider.LLMProvider`, which itself subclasses `Provider`.
    Providers for LLM-generated feedback are more of a plug-and-play variety. This means that the
    base model of your choice can be combined with feedback-specific prompting to generate feedback.

    For example, `relevance` can be run with any base LLM feedback provider. Once the feedback provider
    is instantiated with a base model, the `relevance` function can be called with a prompt and response.

    This means that the base model selected is combined with specific prompting for `relevance` to generate feedback.

    Example:
        ```python
        from trulens.providers.openai import OpenAI
        provider = OpenAI(model_engine="gpt-3.5-turbo")
        provider.relevance(prompt, response)
        ```
    """

    model_config: ClassVar[dict] = dict(arbitrary_types_allowed=True)

    endpoint: Optional[core_endpoint.Endpoint] = None
    """Endpoint supporting this provider.

    Remote API invocations are handled by the endpoint.
    """

    def __init__(self, name: Optional[str] = None, **kwargs):
        super().__init__(name=name, **kwargs)



"""TruLens Core LLM Evaluation Library."""

# WARNING: This file does not follow the no-init aliases import standard.

from importlib.metadata import version
import os

from trulens.core.feedback.feedback import Feedback
from trulens.core.feedback.feedback import SnowflakeFeedback
from trulens.core.feedback.provider import Provider
from trulens.core.schema.feedback import FeedbackMode
from trulens.core.schema.select import Select
from trulens.core.session import Tru
from trulens.core.session import TruSession
from trulens.core.utils import imports as import_utils

# NOTE: workaround for MKL and multiprocessing
# https://github.com/pytorch/csprng/issues/115
os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"

__version__ = version(
    import_utils.safe_importlib_package_name(__package__ or __name__)
)

__all__ = [
    "Tru",  # aliases TruSession for backwards compatibility
    "TruSession",  # main interface
    # app setup
    "FeedbackMode",
    # feedback setup
    "Feedback",
    "SnowflakeFeedback",
    "Select",
    "Provider",
]



"""Make app name and app version fields not nullable.

Revision ID: 7
Revises: 6
Create Date: 2024-08-27 11:09:26
"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "7"
down_revision = "6"
branch_labels = None
depends_on = "6"


def upgrade(config) -> None:
    prefix = config.get_main_option("trulens.table_prefix")

    if prefix is None:
        raise RuntimeError("trulens.table_prefix is not set")

    # ### begin Alembic commands ###
    with op.batch_alter_table(prefix + "apps") as batch_op:
        batch_op.alter_column(
            "app_name",
            existing_type=sa.VARCHAR(length=1024),
            nullable=False,
        )
        batch_op.alter_column(
            "app_version",
            existing_type=sa.VARCHAR(length=1024),
            nullable=False,
        )
    # ### end Alembic commands ###


def downgrade(config) -> None:
    prefix = config.get_main_option("trulens.table_prefix")

    if prefix is None:
        raise RuntimeError("trulens.table_prefix is not set")

    # ### begin Alembic commands ###
    with op.batch_alter_table(prefix + "apps") as batch_op:
        batch_op.alter_column(
            "app_name",
            existing_type=sa.VARCHAR(length=1024),
            nullable=True,
        )
        batch_op.alter_column(
            "app_version",
            existing_type=sa.VARCHAR(length=1024),
            nullable=True,
        )
    # ### end Alembic commands ###
